#ifndef DUNE_CURVEDSURFACEGRID_IMPLICIT_SURFACE_PROJECTION_HH
#define DUNE_CURVEDSURFACEGRID_IMPLICIT_SURFACE_PROJECTION_HH

#include <cmath>
#include <type_traits>

#include <dune/common/parametertree.hh>
#include <dune/common/typeutilities.hh>
#include <dune/curvedsurfacegrid/gridfunctions/analyticgridfunction.hh>

namespace Dune {

/// \brief Closest-point projection to surface given by implicit function
/**
 * Surface S is given by zero-levelset of function F.
 * We assume that F is differentiable in order to evaluate normals and to
 * do an iterative projection.
 **/
template <class F>
class ImplicitSurfaceProjection
{
  using Functor = F;
  using DFunctor = std::decay_t<decltype(derivative(std::declval<F>()))>;

public:
  /// \brief Constructor for a given differentiable function with surface = { x : phi(x) == 0 }
  ImplicitSurfaceProjection (const F& phi, int maxIter = 10)
    : maxIter_(maxIter)
    , phi_(phi)
    , dphi_(derivative(phi_))
  {}

  /// \brief Evaluation of the closest-point projection
  /**
   * Iterative algorithm proposed by Demlow and Dziuk in
   * AN ADAPTIVE FINITE ELEMENT METHOD FOR THE LAPLACE-BELTRAMI OPERATOR ON IMPLICITLY DEFINED SURFACES
   **/
  template <class Domain>
  Domain operator() (const Domain& x0) const
  {
    using std::sqrt;
    using T = typename Domain::value_type;

    Domain x = x0;
    T tol = 10*std::numeric_limits<T>::epsilon();

    auto phi = phi_(x);
    auto sign = phi > 0 ? 1 : phi < 0 ? -1 : 0;
    auto grad_phi = dphi_(x);
    auto grad_phi_norm2 = grad_phi.two_norm2();

    T err = 1;
    for (int i = 0; i < maxIter_ && err > tol; ++i) {
      auto y = x - grad_phi * (phi/grad_phi_norm2);

      auto dist = sign * (y - x0).two_norm();
      auto grad_phi_y = dphi_(y);

      auto& dx = grad_phi_y;
      dx *= -dist/grad_phi_y.two_norm();
      x = x0 + dx;

      phi = phi_(x);
      grad_phi = dphi_(x);
      grad_phi_norm2 = grad_phi.two_norm2();
      err = sqrt(phi*phi / grad_phi_norm2 + (grad_phi / grad_phi_norm2 - dx / dx.two_norm()).two_norm2());
    }

    return x;
  }

  /// \brief Normal vector given by grad(F)/|grad(F)|
  template <class Domain>
  Domain normal (const Domain& x0) const
  {
    auto grad_phi = dphi_(x0);
    return grad_phi / grad_phi.two_norm();
  }

private:
  int maxIter_;
  Functor phi_;
  DFunctor dphi_;
};


/// \brief Closest-point projection to surface given by implicit function using a simple projection algorithm
/**
 * Surface S is given by zero-levelset of function F.
 * We assume that F is differentiable in order to evaluate normals and to
 * do an iterative projection.
 **/
template <class F>
class SimpleImplicitSurfaceProjection
{
  using Functor = F;
  using DFunctor = std::decay_t<decltype(derivative(std::declval<F>()))>;

public:
  /// \brief Constructor for a given differentiable function with surface = { x : phi(x) == 0 }
  SimpleImplicitSurfaceProjection (const F& phi, int maxIter = 10)
    : maxIter_(maxIter)
    , phi_(phi)
    , dphi_(derivative(phi_))
  {}

  /// \brief Evaluation of the closest-point projection
  /**
   * Simple iterative algorithm proposed by Ingo Nitschke in
   * Diskretes Äußeres Kalkül (DEC) auf Oberflächen ohne Rand
   **/
  template <class Domain>
  Domain operator() (Domain x) const
  {
    using std::sqrt;
    using T = typename Domain::value_type;

    T tol = 10*std::numeric_limits<T>::epsilon();

    for (int i = 0; i < maxIter_; ++i) {
      auto phi_x = phi_(x);
      auto grad_phi_x = dphi_(x);
      auto normalize = phi_x / grad_phi_x.two_norm2();

      if (sqrt(phi_x * normalize) < tol)
        break;

      x -= grad_phi_x * normalize;
    }

    return x;
  }

  /// \brief Normal vector given by grad(F)/|grad(F)|
  template <class Domain>
  Domain normal (const Domain& x0) const
  {
    auto grad_phi = dphi_(x0);
    return grad_phi / grad_phi.two_norm();
  }

private:
  int maxIter_;
  Functor phi_;
  DFunctor dphi_;
};


/// \brief Construct a grid function representing a projection to an implicit surface
template <class Grid, class F>
auto implicitSurfaceGridFunction (const F& phi, int maxIter = 10)
{
  return analyticGridFunction<Grid>(ImplicitSurfaceProjection<F>{phi, maxIter});
}

/// \brief Construct a grid function representing a projection to an implicit surface by
/// additionally giving the projection implementation as template
template <class Grid, template <class> class Projection, class F>
auto implicitSurfaceGridFunction (const F& phi, int maxIter = 10)
{
  return analyticGridFunction<Grid>(Projection<F>{phi, maxIter});
}

} // end namespace Dune

#endif // DUNE_CURVEDSURFACEGRID_IMPLICIT_SURFACE_PROJECTION_HH
